<?php

namespace topthink\tiktoken;

use InvalidArgumentException;
use think\helper\Str;
use think\Manager;

class TikToken extends Manager
{
    private const ENCODINGS = [
        'r50k_base'   => [
            'vocab' => 'r50k_base.tiktoken',
            'pat'   => '/\'s|\'t|\'re|\'ve|\'m|\'ll|\'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+/u',
        ],
        'p50k_base'   => [
            'vocab' => 'p50k_base.tiktoken',
            'pat'   => '/\'s|\'t|\'re|\'ve|\'m|\'ll|\'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+/u',
        ],
        'p50k_edit'   => [
            'vocab' => 'p50k_base.tiktoken',
            'pat'   => '/\'s|\'t|\'re|\'ve|\'m|\'ll|\'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+/u',
        ],
        'cl100k_base' => [
            'vocab' => 'cl100k_base.tiktoken',
            'pat'   => '/(?i:\'s|\'t|\'re|\'ve|\'m|\'ll|\'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+/u',
        ],
    ];

    protected $vocabs = [];

    public function nums($model, $messages, $tools = [])
    {
        if (Str::startsWith($model, 'gpt-3.5-turbo-0301')) {
            $perMessage = 4;
            $perName    = -1;
        } elseif (Str::startsWith($model, ['gpt-3.5-turbo', 'gpt-4'])) {
            $perMessage = 3;
            $perName    = 1;
        } else {
            throw new InvalidArgumentException(sprintf('Unknown model name: %s', $model));
        }

        $encoder = $this->getEncoder('cl100k_base');

        if (is_string($messages)) {
            return count($encoder->encode($messages));
        }

        $nums = 0;

        foreach ($messages as $message) {
            $nums += $perMessage;
            foreach ($message as $key => $value) {
                if ($key == 'tool_calls') {
                    foreach ($value as $call) {
                        foreach ($call as $cKey => $cValue) {
                            $nums += count($encoder->encode($cKey));
                            if ($cKey == 'function') {
                                foreach ($cValue as $fKey => $fValue) {
                                    $nums += count($encoder->encode($fKey));
                                    $nums += count($encoder->encode($fValue));
                                }
                            } else {
                                $nums += count($encoder->encode($cValue));
                            }
                        }
                    }
                } else {
                    # TODO: The current token calculation method for the image type is not implemented,
                    #  which need to download the image and then get the resolution for calculation,
                    #  and will increase the request delay
                    if (is_array($value)) {
                        $text = '';
                        foreach ($value as $v) {
                            if (is_array($v) && $v['type'] == 'text') {
                                $text .= $v['text'];
                            }
                        }
                        $value = $text;
                    }
                    if (is_string($value)) {
                        $nums += count($encoder->encode($value));
                    }
                }

                if ($key == 'name') {
                    $nums += $perName;
                }
            }
        }

        $nums += 3;

        if (!empty($tools)) {
            foreach ($tools as $tool) {
                $nums += count($encoder->encode('type'));
                $nums += count($encoder->encode('function'));

                $function = $tool['function'];

                $nums += count($encoder->encode('name'));
                $nums += count($encoder->encode($function['name']));
                $nums += count($encoder->encode('description'));
                $nums += count($encoder->encode($function['description']));

                $parameters = $function['parameters'] ?? null;
                if (!empty($parameters)) {
                    $nums += count($encoder->encode('parameters'));
                    if (isset($parameters['title'])) {
                        $nums += count($encoder->encode('title'));
                        $nums += count($encoder->encode($parameters['title']));
                    }
                    $nums += count($encoder->encode('type'));
                    $nums += count($encoder->encode($parameters['type']));
                    if (isset($parameters['properties'])) {
                        $nums += count($encoder->encode('properties'));
                        foreach ($parameters['properties'] as $key => $value) {
                            $nums += count($encoder->encode($key));
                            foreach ($value as $k => $v) {
                                $nums += count($encoder->encode($k));
                                if ($k == 'enum') {
                                    foreach ($v as $enum) {
                                        $nums += 3;
                                        $nums += count($encoder->encode($enum));
                                    }
                                } else {
                                    $nums += count($encoder->encode($k));
                                    $nums += count($encoder->encode((string) $v));
                                }
                            }
                        }
                    }
                    if (isset($parameters['required'])) {
                        $nums += count($encoder->encode('required'));
                        foreach ($parameters['required'] as $required) {
                            $nums += 3;
                            $nums += count($encoder->encode($required));
                        }
                    }
                }
            }
        }

        return $nums;
    }

    public function getEncoder($name): Encoder
    {
        return $this->driver($name);
    }

    protected function createDriver(string $name)
    {
        if (!isset(self::ENCODINGS[$name])) {
            throw new InvalidArgumentException(sprintf('Unknown encoding: %s', $name));
        }

        $options = self::ENCODINGS[$name];

        return new Encoder(
            $name,
            $this->getVocab($name),
            $options['pat'],
        );
    }

    protected function getVocab(string $name): Vocab
    {
        if (isset($this->vocabs[$name])) {
            return $this->vocabs[$name];
        }

        return $this->vocabs[$name] = Vocab::fromUri(__DIR__ . '/encodings/' . self::ENCODINGS[$name]['vocab']);
    }

    public function getDefaultDriver()
    {
        return null;
    }
}
